/*
 *	Blue book ethernet sockets.
 *
 *	Authors:	Alan Cox <iialan@iifeak.swan.ac.uk>
 *	
 *	Again this doesn't work like the old SOCK_PACKET as it
 *	is trying to be more streams related. Once the streams
 *	support works that will be the preferred access method.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 * 
 */
 
#include <linux/types.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <asm/segment.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include "sock.h"
#include "sockgeneric.h"
#include <linux/if_nit.h>

#ifdef CONFIG_DIX

/*
 *	Things you do differently with an ethernet socket.
 */
 
static int ether_bind(struct socket *sock, struct sockaddr *umyaddr, int sockaddr_len)
{
	int err;
	struct sock *sk=(struct sock *)sock->data;
	struct device *dev;
	unsigned long flags;
	
	if(sk->opt.generic.bound)
		return -EISCONN;
	err=gs_bind(sock,umyaddr,sockaddr_len);		/* Bind takes a device */
	if(err)
		return err;
		
	sk->opt.generic.bound=1;
	save_flags(flags);				/* Avoid any device deactivation races (can't occur anyway but) */
	cli();
	dev=dev_get(sk->opt.generic.device);
	if(dev==NULL)
	{
		restore_flags(flags);
		sk->opt.generic.bound=0;
		return -ENODEV;
	}
	if(dev->type!=ARPHRD_ETHER)
	{
		restore_flags(flags);
		sk->opt.generic.bound=0;
		return -EPROTONOSUPPORT;		/* Driver given isn't ethernet */
	}
	sk->opt.generic.protocol=protocol_clone(&proto_generic);
	if(sk->opt.generic.protocol==0)
	{
		restore_flags(flags);
		sk->opt.generic.bound=0;
		return -ENOBUFS;
	}
	err=protocol_bind(dev->mac_protocol, sk->opt.generic.protocol, sk->num, 0);
	restore_flags(flags);
	if(err)
	{
		protocol_delete(sk->opt.generic.protocol);
		sk->opt.generic.bound=0;
		return err;
	}
	sk->opt.generic.lower=dev->mac_protocol;
	return 0;
}

static int ether_connect(struct socket *sock, struct sockaddr *addr, int addr_len, int nonblock)
{
	struct sockaddr_ether *se=(struct sockaddr_ether *)addr;
	struct sock *sk=(struct sock *)sock->data;
	if(sk->opt.generic.bound==0)
		return -EINVAL;
	if(addr_len!=sizeof(*se))
		return -EINVAL;
	if(se->sether_family!=AF_DIX)
		return -EAFNOSUPPORT;
	memcpy(sk->opt.generic.address, se->sether_addr, ETH_ALEN);
	sk->state=TCP_ESTABLISHED;
	return 0;
}

static int ether_sendto(struct socket *sock, void *buff, int len, int nonblock, unsigned flags, struct sockaddr *addr, int addr_len)
{
	struct sockaddr_ether *se=(struct sockaddr_ether *)addr;
	struct sock *sk=(struct sock *)sock->data;
	sk_buff *skb;
	unsigned char *addr_ptr=sk->opt.generic.address;
	int err;
	
	if(len>1500)
		return -EMSGSIZE;
		
	if(addr!=NULL)
	{
		if(addr_len!=sizeof(*se))
			return -EINVAL;
		if(se->sether_family!=AF_DIX)
			return -EAFNOSUPPORT;
		addr_ptr=se->sether_addr;
	}
	else
	{
		if(se==NULL)
			return -ENOTCONN;
	}
	
	if(flags)
		return -EINVAL;
	if(sk->opt.generic.bound==0)
		return -ENOTCONN;

	if(sk->shutdown&SEND_SHUTDOWN)
		return -EPIPE;
				
	skb=sock_alloc_send_skb(sk, len+protocol_size(sk->opt.generic.lower), nonblock, &err);
	if(skb==NULL)
		return err;
	protocol_adjust(skb, sk->opt.generic.lower);
	memcpy_fromfs(skb_put(skb,len),buff,len);	
	sk->opt.generic.lower->output(sk,skb, sk->num, 0, NULL, addr_ptr, NULL, sk->priority);
	return len;
}

static int ether_send(struct socket *sock, void *buff, int len, int nonblock, unsigned flags)
{
	return ether_sendto(sock,buff, len, nonblock, flags, NULL, 0);
}

static int ether_write(struct socket *sock, char *ubuf, int size, int nonblock)
{
	return ether_send(sock,ubuf,size, nonblock,0);
}

/*
 *	Ethernet specific creation rules.
 */
 
static int ether_create(struct socket *sock, int protocol)
{
	struct sock *sk;
	struct protocol *tmp;

	if(!suser())
		return -EPERM;
		
	if(protocol<1514)
		return -EPROTONOSUPPORT;
	
	if(gs_create(sock, protocol)==-1)
		return -EINVAL;

	sk=(struct sock *)sock->data;
	sk->opt.generic.family=AF_DIX;
	
	/*
	 *	Now attach ourselves.
	 */

	sk->opt.generic.protocol=NULL;
	sk->opt.generic.lower=NULL;
	sk->opt.generic.bound=0;	 
	sk->state=TCP_CLOSE;
	return 0;
}


/*
 *	Release a socket. Unbind the link which will free the
 *	cloned gensocket protocol layer, then let the generic
 *	socket releaser do its work.
 */
 
static int ether_release(struct socket *sock, struct socket *peer)
{
	struct sock *sk=(struct sock *)sock->data;
	
	/*
	 *	Free up any bound protocols 
	 */
	if(sk->opt.generic.bound)
		protocol_unbind(sk->opt.generic.lower, sk->opt.generic.protocol, sk->num, 0);
	/*
	 *	Do the generic release job
	 */
	return gs_release(sock,peer);
}
	

struct proto_ops ether_ops=
{
	AF_DIX,
	ether_create,
	gs_dup,
	ether_release,
	ether_bind,
	ether_connect,
	gs_socketpair,
	gs_accept,
	gs_getname,
	gs_read,
	ether_write,
	gs_select,
	gs_ioctl,
	gs_listen,
	ether_send,
	gs_recv,
	ether_sendto,
	gs_recvfrom,
	gs_shutdown,
	gs_setsockopt,
	gs_getsockopt,
	gs_fcntl
};


void dix_init(void)
{
	printk("NET3 Raw ethernet sockets v002\n");
	sock_register(ether_ops.family, &ether_ops);
}


#endif /* CONFIG_DIX */